from sympy.abc import n

from ramanujantools import Matrix


def walk_benchmark(matrix, trajectory, iterations, start):
    Matrix._walk_inner.cache_clear()
    return matrix.walk(trajectory, iterations, start)


def test_walk_pcf_single_parameter_benchmark(benchmark):
    matrix = Matrix([[0, -(n**2)], [1, n + 1]])
    benchmark(walk_benchmark, matrix, {n: 1}, 1000, {n: 1})


def test_walk_2x2_single_parameter_benchmark(benchmark):
    matrix = Matrix([[n**3 - 1, -(n**2)], [n**2 + 17, 1 / (n + 1)]])
    benchmark(walk_benchmark, matrix, {n: 1}, 1000, {n: 1})


def test_walk_3x3_single_parameter_benchmark(benchmark):
    matrix = Matrix(
        [
            [n**3 - 1, -(n**2), 2],
            [n**2 + 17, n + 1, -(n**3)],
            [1 / (n + 2), 12 * n**2 + 13 * n + 14 * n, 19],
        ]
    )
    benchmark(walk_benchmark, matrix, {n: 1}, 1000, {n: 1})


def test_walk_5x5_single_parameter_benchmark(benchmark):
    matrix = Matrix(
        [
            [n**3 - 1, -(n**2), 2, n, 1],
            [n**2 + 17, n + 1, -(n**3), 7, n**3 - 2],
            [1 / (n + 2), 12 * n**2 + 13 * n + 14 * n, 19, 1 / n, n**2],
            [3 * n**3 - 2, n + 1, 1 / (n + 1), 7, n**2 - 1],
            [n, n**2, n**3, 1 - n, -n * (n - 7) + 3],
        ]
    )
    benchmark(walk_benchmark, matrix, {n: 1}, 1000, {n: 1})


def test_walk_5x5_polynomial_single_parameter_benchmark(benchmark):
    matrix = Matrix(
        [
            [n**3 - 1, -(n**2), 2, n, 1],
            [n**2 + 17, n + 1, -(n**3), 7, n**3 - 2],
            [n + 2, 12 * n**2 + 13 * n + 14 * n, 19, 1, n**2],
            [3 * n**3 - 2, n + 1, n + 1, 7, n**2 - 1],
            [n, n**2, n**3, 1 - n, -n * (n - 7) + 3],
        ]
    )
    benchmark(walk_benchmark, matrix, {n: 1}, 1000, {n: 1})


def test_as_companion_3x3_2f2_benchmark(benchmark):
    matrix = Matrix(
        [
            [n**2 * (n + 2), 2 * n**2 * (n + 1), n**3 * (n + 1)],
            [
                n**2 + 6 * n + 2,
                n**3 + 10 * n**2 + 10 * n + 2,
                6 * n**3 + 5 * n**2 - 6 * n - 4,
            ],
            [-2 * n - 1, -(n**2) + n + 1, (3 * n + 2) * (3 * n + 4)],
        ]
    )
    benchmark(Matrix.as_companion, matrix)



import flint
import sympy as sp

FlintContext = flint.fmpz_mpoly_ctx | flint.fmpq_mpoly_ctx
FlintPoly = flint.fmpz_mpoly | flint.fmpq_mpoly


def flint_ctx(symbols: list[sp.Symbol], fmpz: bool) -> FlintContext:
    """
    Creates a FlintContext
    Args:
        symbols: The symbols to be supported by the FlintContext
        fmpz: if True, returns fmpz_mpoly_ctx. Otherwise returns fmpq_mpoly_ctx.
    """
    ctx_type = flint.fmpz_mpoly_ctx if fmpz else flint.fmpq_mpoly_ctx
    return ctx_type.get(
        [str(symbol) for symbol in list(sorted(symbols, key=str))], "lex"
    )


def flint_from_sympy(poly: sp.Expr, ctx: FlintContext) -> FlintPoly:
    """
    Converts a sympy poly to a flint mpoly.
    """

    def coeff_cast(c, fmpz):
        if fmpz:
            return flint.fmpz(c.numerator)
        else:
            return flint.fmpq(c.numerator, c.denominator)

    gens = tuple(sp.Symbol(str(gen)) for gen in ctx.gens())
    sp_poly = sp.Poly(poly, gens)
    mpoly_type = type(ctx.constant(0))
    monom_dict = {
        monom: coeff_cast(coeff, isinstance(ctx, flint.fmpz_mpoly_ctx))
        for monom, coeff in sp_poly.terms()
    }
    return mpoly_type(monom_dict, ctx)


def flint_to_sympy(poly) -> sp.Expr:
    """
    Factors an mpoly polynomial and returns it as a sp.Expr
    """
    gens = poly.context().gens()
    symbols = [sp.Symbol(str(gen)) for gen in gens]
    content, factors = poly.factor()
    p = sp.simplify(content)
    for factor, multiplicity in factors:
        expr = sum(
            coeff * sp.Mul(*[sym**exp for sym, exp in zip(symbols, monom)])
            for monom, coeff in factor.terms()
        )
        p *= expr**multiplicity
    return p



from pytest import raises

import sympy as sp
from sympy.abc import a, b, c, x, y, n

from ramanujantools import Position, Matrix, simplify
from ramanujantools.cmf import CMF, known_cmfs, pFq

a0, a1 = sp.symbols("a:2")
b0, b1 = sp.symbols("b:2")
c0, c1, c2, c3 = sp.symbols("c:4")
x0, x1, x2, x3 = sp.symbols("x:4")
y0, y1, y2, y3 = sp.symbols("y:4")


def test_symbols():
    cmf = known_cmfs.cmf1()
    expected_axes = {x, y}
    expected_parameters = {c0, c1, c2, c3}
    assert expected_axes == cmf.axes()
    assert expected_parameters == cmf.parameters()
    assert set().union(expected_axes, expected_parameters) == cmf.free_symbols()


def test_rank():
    cmf = known_cmfs.hypergeometric_derived_2F1()
    assert 2 == cmf.rank()


def test_dim():
    cmf = known_cmfs.hypergeometric_derived_2F1()
    assert 3 == cmf.dim()


def test_validate_conserving():
    m = Matrix([[x, x + 17], [y * x, y * 3 - x + 5]])
    cmf = CMF(matrices={x: m, y: m}, validate=False)
    with raises(ValueError):
        cmf.validate_conserving()
    with raises(ValueError):
        cmf = CMF(matrices={x: m, y: m}, validate=True)


def test_validate_negative_cache():
    cmf = known_cmfs.cmf1()
    negative_matrices = {axis: cmf.M(axis, sign=False) for axis in cmf.matrices}
    CMF(cmf.matrices, _negative_matrices_cache=negative_matrices, validate=True)
    random_axis = list(negative_matrices.keys())[0]
    negative_matrices[random_axis] += Matrix.eye(cmf.rank())
    with raises(ValueError):
        CMF(cmf.matrices, _negative_matrices_cache=negative_matrices, validate=True)


def test_subs():
    cmf = known_cmfs.var_root_cmf()
    substitution = Position({c0: 1, c1: 2 * c1 - 3})
    assert cmf.subs(substitution) == CMF(
        {axis: matrix.subs(substitution) for axis, matrix in cmf.matrices.items()}
    )


def test_subs_axes_throw():
    cmf = known_cmfs.e()
    substitution = Position({x: 2})
    with raises(ValueError):
        cmf.subs(substitution)


def test_subs_non_linear_shift_throws():
    cmf = known_cmfs.e()
    substitution = Position({x: x**2})
    with raises(ValueError):
        cmf.subs(substitution)


def test_dual_conserving():
    cmf = known_cmfs.cmf2()
    dual_cmf = cmf.dual()
    dual_cmf.validate_conserving()


def test_dual_inverible():
    cmf = known_cmfs.cmf1()
    assert cmf == cmf.dual().dual()


def test_coboundary_eye():
    cmf = known_cmfs.e()
    assert cmf == cmf.coboundary(Matrix.eye(cmf.rank()))


def test_coboundary_inversible():
    cmf = known_cmfs.e()
    U = Matrix([[x, c0 * x], [y, x * y]])
    assert cmf == cmf.coboundary(U).coboundary(U.inverse())


def test_coboundary_is_conserving():
    cmf = pFq(3, 2, 1)
    x0 = sp.Symbol("x0")
    y0 = sp.Symbol("y0")
    U = Matrix([[x0, c0 * y0], [3 * x0 - 2 * y0, 17 * x0**2]])
    co_cmf = cmf.coboundary(U)
    co_cmf.validate_conserving()
    co_cmf.validate_negative_cache()


def test_sub_cmf_zero_trajectory_throws():
    cmf = pFq(2, 1, -1)
    with raises(ValueError):
        cmf.sub_cmf({a0: {x0: 0, x1: 0, y0: 0}})


def test_sub_cmf_missing_axes_throws():
    cmf = pFq(2, 1, -1)
    with raises(ValueError):
        cmf.sub_cmf({a0: {x0: 1, y0: 0}})


def test_sub_cmf_wrong_axes_throws():
    cmf = pFq(2, 1, -1)
    with raises(ValueError):
        cmf.sub_cmf({a0: {x3: 1, y3: 0}})


def test_sub_cmf_linear_dependent_basis_throws():
    cmf = pFq(2, 1, -1)
    with raises(ValueError):
        cmf.sub_cmf(
            {
                a0: {x0: 1, x1: 0, y0: 0},
                a1: {x0: 1, x1: 0, y0: 0},
            }
        )


def test_sub_cmf_trivial():
    cmf = pFq(2, 1, -1)
    basis = {
        a0: {x0: 1, x1: 0, y0: 0},
        a1: {x0: 0, x1: 1, y0: 0},
        b0: {x0: 0, x1: 0, y0: 1},
    }
    expected = CMF(
        matrices={
            a0: cmf.M(x0).subs({x0: a0, x1: a1, y0: b0}),
            a1: cmf.M(x1).subs({x0: a0, x1: a1, y0: b0}),
            b0: cmf.M(y0).subs({x0: a0, x1: a1, y0: b0}),
        }
    )
    assert expected == cmf.sub_cmf(basis)


def test_sub_cmf_conserving():
    cmf = pFq(3, 3, -1)
    basis = {
        a0: {x0: 0, x1: 1, x2: 0, y0: 0, y1: 1, y2: -1},
        a1: {x0: 1, x1: 0, x2: 1, y0: 0, y1: 0, y2: 0},
        b0: {x0: 0, x1: 0, x2: 0, y0: 0, y1: 1, y2: 1},
        b1: {x0: 1, x1: 0, x2: 0, y0: 1, y1: 0, y2: 0},
    }
    sub_cmf = cmf.sub_cmf(basis)
    sub_cmf.validate_conserving()


def test_trajectory_matrix_axis():
    cmf = known_cmfs.e()
    assert cmf.trajectory_matrix({x: 3, y: 0}, {x: x, y: y}).walk(
        {n: 1}, 1, {n: 0}
    ) == simplify(cmf.M(x).walk({x: 1, y: 0}, 3, {x: x, y: y}))
    assert cmf.trajectory_matrix({x: 0, y: 2}, {x: x, y: y}).walk(
        {n: 1}, 1, {n: 0}
    ) == simplify(cmf.M(y).walk({x: 0, y: 1}, 2, {x: x, y: y}))


def test_trajectory_matrix_diagonal():
    cmf = known_cmfs.e()
    trajectory = Position({x: 1, y: 1})
    start = Position({x: x, y: y})
    assert cmf.trajectory_matrix(trajectory, start) == simplify(
        (cmf.M(x) * cmf.M(y)({x: x + 1})).subs(
            cmf.trajectory_substitution(trajectory, start, n)
        )
    )


def test_back_negates_forward():
    cmf = known_cmfs.e()
    assert Matrix.eye(2) == (cmf.M(x, True) * cmf.M(x, False).subs({x: x + 1})).reduce()
    assert Matrix.eye(2) == (cmf.M(x, False) * cmf.M(x, True).subs({x: x - 1})).reduce()
    assert Matrix.eye(2) == (cmf.M(y, True) * cmf.M(y, False).subs({y: y + 1})).reduce()
    assert Matrix.eye(2) == (cmf.M(y, False) * cmf.M(y, True).subs({y: y - 1})).reduce()


def test_trajectory_matrix_negative_axis():
    cmf = known_cmfs.e()
    start = {x: x, y: y}
    assert cmf.trajectory_matrix({x: -3, y: 0}, start).equal_projectively(
        cmf.M(x, False).walk(
            {x: -1, y: 0}, 3, cmf.trajectory_substitution({x: -3, y: 0}, start, n)
        )
    )
    assert cmf.trajectory_matrix({x: 0, y: -2}, start).equal_projectively(
        cmf.M(y, False).walk(
            {x: 0, y: -1}, 2, cmf.trajectory_substitution({x: 0, y: -2}, start, n)
        )
    )


def test_trajectory_matrix_negative():
    cmf = known_cmfs.hypergeometric_derived_2F1()
    expected = (
        cmf.M(a, sign=True)
        * cmf.M(b, sign=False).subs({a: a + 1})
        * cmf.M(b, sign=False).subs({a: a + 1, b: b - 1})
        * cmf.M(c, sign=False).subs({a: a + 1, b: b - 2})
    )
    actual = cmf.trajectory_matrix({a: 1, b: -2, c: -1}, {a: a, b: b, c: c})
    assert actual.walk({n: 1}, 1, {n: 0}).equal_projectively(expected)


def test_trajectory_matrix_rational():
    cmf = known_cmfs.e()
    start = Position({x: sp.Rational(2, 3), y: sp.Rational(1, 4)})
    trajectory = {x: 1, y: 1}
    symbolic_start = cmf.trajectory_substitution(trajectory, start, n)
    expected = (
        cmf.M(x).subs(symbolic_start) * cmf.M(y).subs(symbolic_start + Position({x: 1}))
    ).factor()
    actual = cmf.trajectory_matrix({x: 1, y: 1}, start)
    assert expected == actual


def test_walk_axis():
    cmf = known_cmfs.e()
    start = {x: 1, y: 1}
    assert cmf.walk({x: 1, y: 0}, 17, start) == cmf.M(x).walk({x: 1, y: 0}, 17, start)
    assert cmf.walk({x: 0, y: 1}, 17, start) == cmf.M(y).walk({x: 0, y: 1}, 17, start)


def test_walk_axis_multiplicity():
    cmf = known_cmfs.e()
    start = {x: 1, y: 1}
    trajectory = Position({x: 1, y: 0})
    depth = 4
    assert cmf.walk(depth * trajectory, 1, start) == cmf.M(x).walk(
        trajectory, depth, start
    )


def test_walk_diagonal():
    cmf = known_cmfs.e()
    trajectory = {x: 1, y: 1}
    start = {x: 1, y: 1}
    Mxy = cmf.trajectory_matrix(trajectory, start)
    assert cmf.walk(trajectory, 9, start) == Mxy.walk({n: 1}, 9, {n: 0})


def test_limit_diagonal():
    cmf = known_cmfs.e()
    trajectory = {x: 1, y: 1}
    start = {x: 1, y: 1}
    Mxy = cmf.walk(trajectory, 1, {x: x, y: y})
    assert cmf.limit(trajectory, 17, start) == Mxy.limit(trajectory, 17, start)


def test_limit_vectors():
    cmf = known_cmfs.e()
    trajectory = {x: 1, y: 3}
    depths = [12, 13, 17]
    start = {x: 2, y: 1}
    initial_values = Matrix([[1, 2, 3], [4, 5, 6]])
    final_projection = Matrix([[1, 2], [3, 4], [5, 6]])
    expected = cmf.limit(trajectory, depths, start, initial_values, final_projection)
    for lim in expected:
        lim.initial_values = initial_values
        lim.final_projection = final_projection
    assert expected == cmf.limit(
        trajectory, depths, start, initial_values, final_projection
    )


def test_walk_list():
    cmf = known_cmfs.e()
    trajectory = {x: 2, y: 3}
    start = {x: 5, y: 7}
    iterations = [1, 2, 3, 17, 29]
    assert cmf.walk(trajectory, iterations, start) == [
        cmf.walk(trajectory, i, start) for i in iterations
    ]


def test_trajectory_substitution_throws_on_collision():
    cmf = known_cmfs.e()
    trajectory = Position({x: 1, y: 0})
    start = Position({x: n, y: 0})
    with raises(ValueError):
        cmf.trajectory_substitution(trajectory, start, n)
    assert {x: n + a, y: 0} == cmf.trajectory_substitution(trajectory, start, a)


def test_trajectory_substitution_axis():
    cmf = known_cmfs.e()
    x_axis = {x: 1, y: 0}
    y_axis = {x: 0, y: 1}
    origin = {x: 0, y: 0}
    assert {x: n, y: 0} == cmf.trajectory_substitution(x_axis, origin, n)
    assert {x: 0, y: n} == cmf.trajectory_substitution(y_axis, origin, n)


def test_trajectory_substitution_diagonal():
    cmf = known_cmfs.e()
    trajectory = {x: 1, y: 2}
    start = {x: 3, y: 5}
    assert {x: 3 + n, y: 5 + 2 * n} == cmf.trajectory_substitution(trajectory, start, n)


def test_trajectory_matrix_walk_equivalence():
    cmf = known_cmfs.e()
    iterations = 7
    trajectory = {x: 1, y: 1}
    start = {x: 3, y: 5}
    trajectory_matrix = cmf.trajectory_matrix(trajectory, start)
    assert trajectory_matrix.walk({n: 1}, iterations, {n: 0}) == cmf.walk(
        trajectory, iterations, start
    )


def test_work_non_integer_trajectory_throws():
    cmf = known_cmfs.e()
    with raises(ValueError):
        cmf.work(Position({x: 1, y: 1}), Position({x: 2, y: sp.Rational(5, 2)}))
    with raises(ValueError):
        cmf.work(Position({x: 1, y: n}), Position({x: 1, y: 2 * n}))


def test_work_numeric():
    cmf = known_cmfs.e()
    start = Position({x: 1, y: -1})
    end = Position({x: 17, y: -13})
    trajectory = end - start
    expected = cmf.walk(trajectory, 1, start)
    actual = cmf.work(start, end)
    assert expected == actual


def test_work_symbolic():
    cmf = known_cmfs.e()
    start = Position({x: 1 + x, y: -1 + 7 * y / 2})
    end = Position({x: 17 + x, y: -13 + 7 * y / 2})
    trajectory = end - start
    expected = cmf.walk(trajectory, 1, start)
    actual = cmf.work(start, end)
    assert expected == actual


def test_delta_correctness():
    cmf = known_cmfs.hypergeometric_derived_2F1()
    depth = 100
    trajectory = {a: 1, b: 0, c: 1}
    start = {a: 1, b: 1, c: 1}
    l1, l2 = cmf.limit(trajectory, [depth, 2 * depth], start)
    assert cmf.delta(trajectory, depth, start) == l1.delta(l2.as_float())
    assert cmf.delta(trajectory, depth, start, limit=l2.as_float()) == l1.delta(
        l2.as_float()
    )


def test_blind_delta_sequence_agrees_with_blind_delta():
    cmf = known_cmfs.hypergeometric_derived_2F1()
    depth = 30
    trajectory = {a: 1, b: 2, c: 0}
    start = {a: 2, b: 3, c: 4}
    limit = cmf.limit(trajectory, 2 * depth, start).as_float()
    delta_sequence = cmf.delta_sequence(trajectory, depth, start)
    sequence_of_deltas = [
        cmf.delta(trajectory, i, start, limit=limit) for i in range(1, depth + 1)
    ]
    assert delta_sequence == sequence_of_deltas
    assert len(delta_sequence) == depth



from __future__ import annotations

from typing import Callable

import sympy as sp
import mpmath as mp

from ramanujantools import Matrix


def first_unmatch(a: str, b: str) -> int:
    """
    Returns the index of the biggest digit that does not match in a and b
    """
    size = min(len(a), len(b))
    for i in range(size):
        if a[i] != b[i]:
            return i
    return size


def round_attempt(lower_bound: mp.mpf, upper_bound: mp.mpf) -> str:
    lower = str(lower_bound)
    upper = str(upper_bound)
    up_to = first_unmatch(lower, upper) + 1
    return upper[0:up_to]


def most_round_in_range(num: mp.mpf, err: mp.mpf) -> str:
    return min(round_attempt(num, num + err), round_attempt(num - err, num), key=len)


def coefficients_from_pslq(
    pslq_results: list[int], active_indices: list[int], N: int
) -> list[int]:
    coefficients = [0] * N
    for i in active_indices:
        coefficients[i] = pslq_results.pop(0)
    return coefficients


class Limit:
    r"""
    Represents a mathematical limit of a `walk` operation.

    Contains two matrices for the two last steps of calculation.
    Uses the last step to extract constants, and the previous one to determine precision.
    The diagonal of the 2x2 matrix calculated as $IV \cdot M \cdot FP$ consists of the p and q values
    that define the limit as $p/q$.
    """

    def __init__(
        self,
        current: Matrix,
        previous: Matrix,
        initial_values: Matrix | None = None,
        final_projection: Matrix | None = None,
    ):
        self.current = current
        self.previous = previous
        self.initial_values = (
            initial_values
            or Matrix.hstack(
                Matrix.e(self.N(), 0),
                Matrix.e(self.N(), 1),
            ).transpose()
        )
        self.final_projection = final_projection or Matrix.hstack(
            Matrix.e(self.N(), -1),
            Matrix.e(self.N(), -1),
        )

    def __repr__(self) -> str:
        return f"Limit({self.current}, {self.previous}, {self.initial_values}, {self.final_projection})"

    def __str__(self) -> str:
        return repr(self)

    def __eq__(self, other: Limit) -> bool:
        return (
            self.current == other.current
            and self.previous == other.previous
            and self.initial_values == other.initial_values
            and self.final_projection == other.final_projection
        )

    def N(self) -> int:
        """
        Returns N, the dimension of the square limit matrix.
        """
        return self.current.rows

    def p_vectors(self) -> list[Matrix]:
        """
        Returns the IV row vector and the FP column vector used to calculate p.
        """
        return [self.initial_values.row(0), self.final_projection.col(0)]

    def q_vectors(self) -> list[Matrix]:
        """
        Returns the IV row vector and the FP column vector used to calculate q.
        """
        return [self.initial_values.row(1), self.final_projection.col(1)]

    def p(self, previous=False) -> sp.Rational:
        """
        Returns p, the numerator of the rational limit.

        Args:
            previous: if True, will use `self.previous` matrix to calculate p, else `self.current`. False by default.
        """
        matrix = self.previous if previous else self.current
        row, col = self.p_vectors()
        return (row * matrix * col)[0]

    def q(self, previous=False) -> sp.Rational:
        """
        Returns q, the numerator of the rational limit.

        Args:
            previous: if True, will use `self.previous` matrix to calculate q, else `self.current`. False by default.
        """
        matrix = self.previous if previous else self.current
        row, col = self.q_vectors()
        return (row * matrix * col)[0]

    @property
    def mp(self):
        """
        Returns an mpmath context with the precision set to 15 digits.
        """
        limit_ctx = mp.mp.clone()
        limit_ctx.dps = max(self.precision() * 1.1, 15)
        return limit_ctx

    @staticmethod
    def walk_to_limit(
        iterations: list[int],
        walk_function: Callable[list[int], list[Limit]],
        initial_values: Matrix | None = None,
        final_projection: Matrix | None = None,
    ) -> list[Limit]:
        """
        Utility function that receives a walk function logic and a list of iterations,
        and returns a list of Limit objects for these iterations

        Args:
            iterations: A list of the required iterations to calculate the limits for.
            walk_function: The walk logic function.
            initial_values: The initial values matrix, defaults to $e_1$ and $e_2$.
            final_projection: The final projection matrix, defaults to $e_{-1}$ and $e_{-1}$.
        """
        previous_values = [max(depth - 1, 0) for depth in iterations]
        walk_iterations = sorted(list(set(iterations + previous_values)))
        walk_matrices = walk_function(walk_iterations)

        current_index = 0
        limits = []
        for depth in iterations:
            while depth != walk_iterations[current_index]:
                current_index += 1
            limits.append(
                Limit(
                    walk_matrices[current_index],
                    walk_matrices[current_index - 1],
                    initial_values,
                    final_projection,
                )
            )
        return limits

    def as_rational(self, previous=False) -> sp.Rational:
        r"""
        Returns the limit as a rational number $\frac{p}{q}$.

        Args:
            previous: Will use $M$ = `self.previous` if True, else `self.current`. False by default.
        """
        return sp.Rational(self.p(previous), self.q(previous))

    def precision(self, base: int = 10) -> int:
        """
        Returns the error in 'digits' for the PCF convergence.

        Args:
            base: The numerical base in which to return the precision (by default 10)
        """
        diff = self.as_rational(previous=False) - self.as_rational(previous=True)
        if not diff.is_rational:  # division by 0, etc
            return 0
        if diff == 0:
            return 100  # big enough, this should be infinity
        # extracting real because sometimes log returns a complex with tiny imaginary type due to precision
        digits = -mp.re(mp.log(diff, base))
        return int(mp.floor(digits))

    def as_float(self) -> mp.mpf:
        r"""
        Returns the limit as a floating point number f, such that $m \cdot v = f$, where `m=self` and `v=vector`.
        """
        return self.mp.mpf(self.as_rational())

    def as_rounded_number(self) -> str:
        """
        Same as `as_float`, but also rounds the result to the shortest number possible within the error range.
        """
        return most_round_in_range(self.as_float(), 10 ** -self.precision())

    def delta(self, L: mp.mpf) -> float:
        r"""
        Calculates the irrationality measure $\delta$ defined, as:
        $|\frac{p}{q} - L| = \frac{1}{{q}^{1+\delta}}$
        Args:
            L: $L$
        Returns:
            the delta value as defined above.
        """
        p, q = self.as_rational().as_numer_denom()
        if q == 0:
            return self.mp.mpf("-inf")
        if q == 1:
            return self.mp.mpf("inf")
        return float(-(1 + self.mp.log(self.mp.fabs(L - (p / q)), q)))

    def identify_rational(self, column_index=-1, maxcoeff=1000) -> Matrix | None:
        r"""
        Searches for a vector $c$ to match a column $p$ of the limit matrix,
        such that $c \cdot p \approx 0$,

        This is essentially the same as `self.identify(0, column_index, maxcoeff)`
        Args:
            column_index: The column index that defines $p$. -1 by default (rightmost column).
        Returns:
            A 1-row matrix $c$ such that $c \cdot p \approx 0$.
        """
        pslq_result = self.mp.pslq(self.current.col(column_index), maxcoeff=maxcoeff)
        if pslq_result is None:
            return None
        return Matrix([coefficients_from_pslq(pslq_result, range(self.N()), self.N())])

    def identify(self, L: mp.mpf, column_index=-1, maxcoeff=1000) -> Matrix | None:
        r"""
        Given a constant $L$, searches for two vectors $a, b$
        such that $\frac{a \cdot p}{b \cdot p} \approx L$,
        Where $p$ a column of the limit matrix.

        Args:
            column_index: The column index that defines $p$. -1 by default (rightmost column).
        Returns:
            A 2-row initial values matrix $m$ (made from rows $a$ and $b$),
            such that $\frac{a \cdot p}{b \cdot p} \approx L$.
        """
        if L == 0:
            return self.identify_rational()

        def linear_independent_indices():
            indices = list(range(self.N()))

            def remove_index(pslq_result):
                for index, value in reversed(list(enumerate(pslq_result))):
                    if value != 0:  # found a dependency
                        indices.pop(index)
                        return

            while len(indices) > 1:  # a single index is linear independent vacuously
                integer_sequences = [self.current.col(column_index)[i] for i in indices]
                pslq_result = self.mp.pslq(integer_sequences, maxcoeff=maxcoeff)
                if pslq_result is None:
                    return indices
                remove_index(pslq_result)

            return indices

        used_indices = linear_independent_indices()
        total_indices = len(used_indices)
        integer_sequences = [self.current.col(column_index)[i] for i in used_indices]
        to_identify = integer_sequences + [p * L for p in integer_sequences]
        pslq_result = self.mp.pslq(to_identify, maxcoeff=maxcoeff, maxsteps=maxcoeff)
        if pslq_result is None:
            return None
        # We normalize the sign so that the first coefficient is positive,
        # to avoid ambiguity as results are equivalent up to the sign.
        numerator_coeffs = coefficients_from_pslq(
            pslq_result[:total_indices], used_indices, self.N()
        )
        numerator = Matrix([numerator_coeffs])
        denominator_coeffs = coefficients_from_pslq(
            pslq_result[total_indices:], used_indices, self.N()
        )
        denominator = Matrix([denominator_coeffs])
        sign = sp.sign(numerator[0]) or 1
        initial_values = Matrix.vstack(sign * numerator, -sign * denominator)
        self.initial_values = initial_values
        self.final_projection = Matrix.hstack(
            Matrix.e(self.N(), column_index),
            Matrix.e(self.N(), column_index),
        )
        return initial_values



from __future__ import annotations

import sympy as sp


class Position(dict[sp.Symbol, sp.Expr]):
    r"""
    Represents a Position (or a trajectory).

    Inherits from dict and adds algebraic utilities.

    Example:
        >>> p = Position({x: 1, y: 2, z: 3})
        >>> p
        Position({x: 1, y: 2, z: 3})
        >>> 2 * p
        Position({x: 2, y: 4, z: 6})
        >>> trajectory = Position({x: 3, y: -1, z: 2})
        >>> p += 4 * trajectory
        Position({x: 13, y: -2, z: 11})
    """

    def __repr__(self):
        return f"Position({super().__repr__()})"

    def __iadd__(self, other: dict):
        for key in other:
            self[key] = self.get(key, 0) + other[key]
        return self

    def __add__(self, other: dict):
        result = self.copy()
        result += other
        return result

    def __isub__(self, other: dict):
        self += -other
        return self

    def __sub__(self, other: dict):
        result = self.copy()
        result -= other
        return result

    def __mul__(self, coefficient: int):
        return Position({key: coefficient * value for key, value in self.items()})

    def __rmul__(self, coefficient: int):
        return self * coefficient

    def __neg__(self):
        return -1 * self

    def __hash__(self):
        return hash(frozenset(self.items()))

    def copy(self):
        return Position(super().copy())

    def longest(self):
        r"""
        Returns the longest element of `this`, i.e, the largest (abs) value in it.
        This is the $ L_\infty $ norm of the vector.
        """
        return max([abs(value) for value in self.values()], default=0)

    def shortest(self):
        r"""
        Returns the shortest nonzero element of `this`, i.e, the smallest (abs) nonzero value in it.
        """
        return min([abs(value) for value in self.values() if value != 0], default=0)

    def signs(self):
        r"""
        Returns the sign in each direction of this
        """
        return Position({key: int(sp.sign(value)) for key, value in self.items()})

    def is_polynomial(self) -> bool:
        """
        Returns true iff all position elements are polynomial
        """
        for element in self.values():
            if not all(
                [
                    c.is_Integer
                    for c in sp.simplify(element).as_coefficients_dict().values()
                ]
            ):
                return False
        return True

    def denominator_lcm(self) -> int:
        r"""
        Returns the lcm of all denominators
        """
        return int(
            sp.lcm(
                [sp.simplify(element).as_numer_denom()[1] for element in self.values()]
            )
        )

    def is_integer(self) -> bool:
        """
        Returns true iff all position elements are numerical
        """
        return all(sp.simplify(element).is_Integer for element in self.values())

    def free_symbols(self) -> set:
        symbols = set()
        for value in self.values():
            symbols = symbols.union((sp.S(0) + value).free_symbols)
        return symbols

    def sorted(self) -> Position:
        return Position({key: self[key] for key in sorted(self.keys(), key=str)})

    @staticmethod
    def from_list(values: list[sp.Expr], symbol: str) -> Position:
        """
        Constructs a Position object with incrementing symbols from a list.
        Example:
            >>> Position.from_list([1, 2, 3], "x")
            Position({x0: 1, x1: 2, x2: 3})
        """
        amount = len(values)
        symbols = sp.symbols(f"{symbol}:{amount}")
        return Position({symbols[i]: values[i] for i in range(amount)})



import functools
import inspect
from typing import (
    Any,
    Callable,
    Union,
    TypeVar,
    TypeAlias,
    get_origin,
    get_args,
    ForwardRef,
    _eval_type,
)
from types import UnionType

X = TypeVar("X")
Batchable: TypeAlias = X | list[X]


def batched(
    arg_name: str,
) -> Callable:
    """
    Decorator that makes a function accept either a scalar or a list for the
    specified argument name. The decorated function must accept a List[X] for
    that argument and return a List[X].

    The decorator reads the type hint of the argument and validates inputs
    accordingly, raising TypeError on mismatches.

    Parameters:
        arg_name: The name of the argument to batchify.

    Returns:
        A decorator that wraps the function accordingly.
    """

    def decorator(func: Callable) -> Callable:
        raw_annot = func.__annotations__.get(arg_name)
        if raw_annot is None:
            raise TypeError(f"No annotation for argument '{arg_name}'")

        # Now evaluate just this annotation safely
        globalns = func.__globals__
        localns = {}
        if isinstance(raw_annot, str):
            raw_annot = ForwardRef(raw_annot)
        annot = _eval_type(raw_annot, globalns, localns, recursive_guard=set())
        origin = get_origin(annot)
        args = get_args(annot)

        if origin not in (Union, UnionType) or len(args) != 2:
            raise TypeError(
                f"Parameter '{arg_name}' annotation must be X | List[X], got {annot}"
            )

        scalar_type = args[0]

        sig = inspect.signature(func)
        param = sig.parameters.get(arg_name)
        if param is None:
            raise ValueError(f"Function {func.__name__} has no parameter '{arg_name}'")

        # Assume list_type is List[scalar_type], no checks here as per your request

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Batchable:
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()
            val = bound_args.arguments.get(arg_name)

            if val is None:
                raise ValueError(f"Argument '{arg_name}' must be provided")

            if isinstance(val, list):
                if not all(isinstance(el, scalar_type) for el in val):
                    raise TypeError(
                        f"All elements of argument '{arg_name}' must be of type Batchable[{scalar_type.__name__}] "
                        f"(i.e., {scalar_type.__name__} or list[{scalar_type.__name__}]), not {type(val).__name__}"
                    )
                return func(*args, **kwargs)
            else:
                if not isinstance(val, scalar_type):
                    raise TypeError(
                        f"Argument '{arg_name}' must be of type Batchable[{scalar_type.__name__}] "
                        f"(i.e., {scalar_type.__name__} or list[{scalar_type.__name__}]), not {type(val).__name__}"
                    )

                new_args = list(bound_args.args)
                new_kwargs = dict(bound_args.kwargs)

                param_pos = list(sig.parameters).index(arg_name)
                if param_pos < len(new_args):
                    new_args[param_pos] = [val]
                else:
                    new_kwargs[arg_name] = [val]

                result = func(*new_args, **new_kwargs)
                return result[0]

        return wrapper

    return decorator



from pytest import approx

import sympy as sp
from sympy.abc import n, z

from ramanujantools import Position, Matrix
from ramanujantools.cmf import MeijerG


def test_conserving():
    for _p in range(1, 3):
        for _q in range(1, 3):
            for _n in range(_p):
                for _m in range(_q):
                    MeijerG(_m, _n, _p, _q, z).validate_conserving()


def test_gamma():
    cmf = MeijerG(3, 2, 2, 3, 1)
    a0, a1 = sp.symbols("a:2")
    b0, b1, b2 = sp.symbols("b:3")
    start = Position({a0: 0, a1: 0, b0: 0, b1: 0, b2: 0})
    trajectory = Position({a0: -1, a1: -1, b0: 1, b1: 1, b2: 1})
    m = cmf.trajectory_matrix(trajectory, start)

    mm = m.inverse().transpose().as_polynomial()
    limit = mm.limit({n: 1}, 200, {n: 0})
    limit.initial_values = Matrix([[1, 1, 0], [1, 1, 1]])

    assert limit.as_float() == approx(limit.mp.euler)



import sympy as sp
from sympy.abc import x, y

from ramanujantools.flint_core import flint_ctx, FlintRational


def flintify(expr: sp.Expr, symbols: list = None, fmpz=True) -> FlintRational:
    ctx = flint_ctx(symbols or list(expr.free_symbols), fmpz)
    return FlintRational.from_sympy(expr, ctx)


def test_from_sympy():
    expression = (x + y - 3) / (x**2 - y)
    ctx = flint_ctx(["x", "y"], fmpz=True)
    _x, _y = ctx.gens()
    expected = FlintRational(_x + _y - 3, _x**2 - _y, ctx)
    assert expected == flintify(expression)


def test_add():
    expr1 = x / y
    expr2 = (3 + x) / (7 - y) + y
    assert flintify(expr1 + expr2) == flintify(expr1) + flintify(expr2)


def test_sub():
    expr1 = x / y
    expr2 = (3 + x) / (7 - y) + y
    assert flintify(expr1 - expr2) == flintify(expr1) - flintify(expr2)


def test_mul():
    expr1 = x / y
    expr2 = (3 + x) / (7 - y) + y
    assert flintify(expr1 * expr2) == flintify(expr1) * flintify(expr2)


def test_mul_scalar():
    expr = x / y
    assert flintify(expr * 7) == flintify(expr) * 7


def test_rmul_scalar():
    expr = x / y
    assert flintify(7 * expr) == 7 * flintify(expr)


def test_div_scalar():
    expr = x / y
    assert flintify(expr / 7) == flintify(expr) / 7


def test_rdiv_scalar():
    expr = x / y
    assert flintify(7 / expr) == 7 / flintify(expr)


def test_eq():
    expr = x / y
    assert expr == expr / 2 + 3 - 3 + expr / 2


def test_factor():
    expected = (x + y) * (x**2 + y**2) * (y - 3) / ((x + 17) * (y - 15) * (x * y + 1))
    assert expected == flintify(expected.expand()).factor()


def test_subs_integer():
    expr = (x**2 + 3 * y / 2 - sp.Rational(5, 4)) / (x * y + y**2 / 3)
    subs = {x: 3, y: x}
    assert flintify(expr.subs(subs), symbols=[x, y]) == flintify(expr).subs(subs)


def test_subs_rational():
    expr = (x**2 + 3 * y / 2 - sp.Rational(5, 4)) / (x * y + y**2 / 3)
    subs = {x: sp.Rational(2, 7), y: (x + 5) / 4}
    assert flintify(expr.subs(subs), symbols=[x, y], fmpz=False) == flintify(
        expr, fmpz=False
    ).subs(subs)



import sympy
from sympy import Poly, Symbol
from sympy.abc import x

from ramanujantools import Matrix, GenericPolynomial, simplify


class CoboundarySolver:
    r"""
    Used to find coboundary relations, namely given to parametrized matrices $m_1(x), m_2(x)$,
    is there a solution $m(x)$ to
    $m1(x) * m(x+1) = m(x) * m2(x)$.
    """

    @staticmethod
    def solve_polynomial_matrix(
        matrix: Matrix, symbol: Symbol, variables: list[Symbol]
    ) -> dict:
        r"""
        Given a polynomial matrix in symbol, such that the coefficients are over 'variables',
        try to find an assignment for this variables so that matrix(assignment) == 0.
        Return this assignment (possibly empty if there is no such assignment, or if the matrix is already zero).

        Note: Might not work if there are more variables than the given list in the parameters.
        """
        # TODO 1 : move this method to a more general location, so other places can use it, also
        # TODO 2 : consider polynomial matrices in several variables as well
        equations = (
            Poly(matrix[0, 0], symbol).all_coeffs()
            + Poly(matrix[1, 0], symbol).all_coeffs()
            + Poly(matrix[0, 1], symbol).all_coeffs()
            + Poly(matrix[1, 1], symbol).all_coeffs()
        )
        return sympy.solve(equations, variables)

    @staticmethod
    def find_coboundary(
        m1: Matrix, m2: Matrix, max_deg: int, symbol: Symbol = x
    ) -> tuple[Matrix, list[Symbol]] | None:
        r"""
        Given two parametrized matrices $m_1(s), m_2(s)$ over the given symbol,
        look for a parametrized polynomial matrix m(s) such that
                $ m_1(s) * m(s+1) = m(s) * m_2(s) $.
        The max degree for the polynomials in m(s) is the given max_deg.
        If there is such a matrix, then returns a tuple of the matrix and the free variables,
        otherwise, returns None.

        Note 1: A solution to the equation above is a vector space, so theoretically it should always
        contain the m(x)=0 solution, but as preparation for a more general solver, I add the None possibility.

        Note 2: Will probably not work if $m_1$ or $m_2$ depend on a variable different from the given symbol.
        """

        f11, f11_ = GenericPolynomial.of_degree(max_deg, "f11_", symbol)
        f12, f12_ = GenericPolynomial.of_degree(max_deg, "f12_", symbol)
        f21, f21_ = GenericPolynomial.of_degree(max_deg, "f21_", symbol)
        f22, f22_ = GenericPolynomial.of_degree(max_deg, "f22_", symbol)
        all_vars = f11_ + f12_ + f21_ + f22_

        # use '.expr' instead of just the polynomials, because calling subs(x,x+1) on polynomials in sympy
        # changes their "base" symbol to x+1 and then raises errors when trying to add them to polynomials
        # with base variable x.
        generic_m = Matrix([[f11.expr, f12.expr], [f21.expr, f22.expr]])

        equations = simplify(m1 * generic_m({symbol: symbol + 1}) - generic_m * m2)
        assignment = CoboundarySolver.solve_polynomial_matrix(
            equations, symbol, all_vars
        )

        if len(assignment) == 0:
            return None

        m = generic_m.subs(assignment)
        vars_left = [v for v in all_vars if v not in assignment.keys()]

        return m, vars_left

    @staticmethod
    def check_unique_solution(
        matrix: Matrix, variables: list[Symbol]
    ) -> tuple[Matrix, list[Symbol]]:
        r"""
        Assuming matrix is linearly dependent on the given variables, in case there is only 1 variable v in vars,
        check if matrix = v*matrix' and if so return (matrix', []).
        Otherwise, return (matrix,vars).
        """
        if len(variables) == 1:
            v = variables[0]
            if matrix.subs({v: 0}) == Matrix(
                [[0, 0], [0, 0]]
            ):  # the matrix m is linear in the variables
                return matrix.subs({v: 1}), []

        return matrix, variables



from __future__ import annotations
from functools import lru_cache

import itertools

import sympy as sp
from sympy.abc import n

from ramanujantools import Position, Matrix, Limit
from ramanujantools.flint_core import (
    NumericMatrix,
    SymbolicMatrix,
    FlintContext,
    flint_ctx,
)

from ramanujantools.utils import batched, Batchable


class CMF:
    r"""
    Represents a Conservative Matrix Field (CMF).

    A CMF is defined by a set of axes and their relevant matrices that satisfy the perservation quality:
    for every two axes x and y, $Mx(x, y) \cdot My(x+1, y) = My(x, y) \cdot Mx(x, y+1)$.
    """

    def __init__(
        self,
        matrices: dict[sp.Symbol, Matrix],
        _negative_matrices_cache: dict[sp.Symbol, Matrix] | None = None,
        validate: bool = True,
    ):
        """
        Initializes a CMF with `Mx` and `My` matrices.

        Args:
            matrices: The CMF matrices as a dict,
                where the keys are the axes of the CMF and their values are the corresponding matrices.

        Raises:
            ValueError: if one of the matrices contain the symbol n, which is recserved for PCF conversions.
        """
        self.matrices = matrices
        self._negative_matrices_cache = _negative_matrices_cache or {}
        if n in self.matrices.keys():
            raise ValueError(
                "Do not use symbol n as an axis, it's reserved for companion conversions"
            )
        self.validate_matrices_same_dimension()
        if validate:
            self.validate_conserving()
            self.validate_negative_cache()

    def __hash__(self) -> int:
        return hash(frozenset(self.matrices.items()))

    def __eq__(self, other) -> bool:
        return self.matrices == other.matrices

    def __repr__(self) -> str:
        return f"CMF({self.matrices})"

    def _latex(self, printer) -> str:
        lines = []
        for axis in sorted(self.axes(), key=str):
            lines.append(
                f"{printer.doprint(axis)} \\mapsto {printer.doprint(self.M(axis))}"
            )
        return r"\begin{array}{l}" + r"\\ ".join(lines) + r"\end{array}"

    def _repr_latex_(self) -> str:
        return rf"{sp.latex(self)}"

    def __getstate__(self):
        return self.matrices

    def __setstate__(self, state):
        self.matrices = state

    def _are_conserving(
        self,
        x: sp.Symbol,
        y: sp.Symbol,
    ) -> bool:
        ctx = self.ctx(Position({x: 0, y: 0}))
        Mx = SymbolicMatrix.from_sympy(self.M(x, True), ctx)
        My = SymbolicMatrix.from_sympy(self.M(y, True), ctx)
        Mxy = Mx * My.subs({x: x + 1})
        Myx = My * Mx.subs({y: y + 1})
        return Mxy == Myx

    def validate_conserving(self) -> None:
        """
        Validates that all of the matrices of the CMF are conserving.
        Raises:
            ValueError: if two matrices or more are not conserving.
        """
        for x, y in itertools.combinations(self.matrices.keys(), 2):
            if not self._are_conserving(x, y):
                raise ValueError(f"M({x}) and M({y}) matrices are not conserving!")

    def validate_matrices_same_dimension(self) -> None:
        """
        Validates that all of the matrices of the CMF have the same dimensions.
        Raises:
            ValueError: in case the matrices are not conserving.
        """
        expected_N = self.rank()
        for symbol, matrix in self.matrices.items():
            if not expected_N == matrix.rows and expected_N == matrix.cols:
                raise ValueError(
                    f"M({symbol}) is of dimension {matrix.rows}x{matrix.cols}, expected {expected_N}x{expected_N}"
                )

    def validate_negative_cache(self) -> None:
        """
        Validates all the matrices in the negative matrices cache.
        """
        ctx = self.ctx()
        for key, inverse_matrix in self._negative_matrices_cache.items():
            M = SymbolicMatrix.from_sympy(self.M(key, True), ctx)
            m_inv = SymbolicMatrix.from_sympy(inverse_matrix, ctx)
            if not (
                SymbolicMatrix.eye(self.rank(), ctx) == (M.subs({key: key - 1}) * m_inv)
            ):
                raise ValueError(
                    f"The negative matrix cache is corrupted for axis {key}"
                )

    def M(self, axis: sp.Symbol, sign: bool = True) -> Matrix:
        """
        Returns the axis matrix for a given axis.

        If sign is negative, returns the matrix corresponding a step back.
        Note that we do not reduce M because it might impair the conservative property.
        """
        if sign:
            return self.matrices[axis]
        else:
            if axis not in self._negative_matrices_cache:
                self._negative_matrices_cache[axis] = self.matrices[axis].inverse()(
                    {axis: axis - 1}
                )
            return self._negative_matrices_cache[axis]

    def axes(self) -> set[sp.Symbol]:
        """
        Returns the symbols of all axes of the CMF.
        """
        return set(self.matrices.keys())

    def parameters(self) -> set[sp.Symbol]:
        """
        Returns all (non-axis) symbolic parameters of the CMF.
        """
        return self.free_symbols() - self.axes()

    def free_symbols(self) -> set[sp.Symbol]:
        """
        Returns all symbolic variables of the CMF, both axes and parameters.
        """
        return set.union(
            *list(map(lambda matrix: matrix.free_symbols, self.matrices.values()))
        )

    def dim(self) -> int:
        """
        Returns the dimension of the CMF,
        which is defined as the amount of axes of the CMF.
        """
        return len(self.axes())

    def rank(self) -> int:
        """
        Returns the rank of the CMF.
        """
        random_matrix = list(self.matrices.values())[0]
        return random_matrix.rows

    def _validate_axes_substitutions(self, substitutions: Position) -> None:
        if (
            axes_substitutions := self.axes().intersection(substitutions.keys())
            != set()
        ):
            raise ValueError(
                f"Cannot substitute axis parameters! got: {axes_substitutions}"
            )

    def subs(self, substitutions: Position) -> CMF:
        """
        Returns a new CMF with substituted matrices.
        """
        self._validate_axes_substitutions(substitutions)
        return CMF(
            matrices={
                symbol: matrix.subs(substitutions)
                for symbol, matrix in self.matrices.items()
            },
            _negative_matrices_cache={
                symbol: matrix.subs(substitutions)
                for symbol, matrix in self._negative_matrices_cache.items()
            },
            validate=False,
        )

    def simplify(self):
        """
        Returns a new CMF with simplified Mx and My
        """
        return CMF(
            matrices={
                symbol: matrix.simplify() for symbol, matrix in self.matrices.items()
            },
            _negative_matrices_cache={
                symbol: matrix.simplify()
                for symbol, matrix in self._negative_matrices_cache.items()
            },
            validate=False,
        )

    def dual(self) -> CMF:
        """
        Returns the dual CMF which is defined with inverse-transpose matrices.
        """
        return CMF(
            {axis: self.M(axis).inverse().transpose() for axis in self.axes()},
            validate=False,
        )

    def coboundary(self, coboundary: Matrix) -> CMF:
        """
        Returns the CMF up to a coboundary matrix U.
        """

        matrices = {
            axis: matrix.coboundary(coboundary, axis, sign=True)
            for axis, matrix in self.matrices.items()
        }
        negative_matrices = {
            axis: matrix.coboundary(coboundary, axis, sign=False)
            for axis, matrix in self._negative_matrices_cache.items()
        }
        return CMF(matrices, negative_matrices, validate=False)

    def sub_cmf(self, basis: dict[sp.Symbol, Position]) -> CMF:
        """
        Returns a sub-CMF defined by a basis of trajectories.
        Each axis in the new CMF corresponds to a trajectory in the basis.
        Validates that the basis is linearly independent.
        Args:
            basis: dict mapping new axis symbols to Position objects (trajectories).
        Returns:
            A new CMF with axes as basis keys and matrices as trajectory matrices.
        Raises:
            ValueError: if the basis is not linearly independent.
        """
        for axis, trajectory in basis.items():
            if set(trajectory.keys()) != self.axes():
                raise ValueError(
                    f"Trajectory for axis {axis} is missing axes {set(self.axes()) - set(trajectory.keys())}"
                )

        basis_matrix = Matrix(
            [
                [basis[new_axis][axis] for new_axis in basis.keys()]
                for axis in self.axes()
            ]
        ).transpose()
        if basis_matrix.rank() < len(basis):
            raise ValueError("Given basis is not linearly independent!")
        start = sum(
            [axis * Position(trajectory) for axis, trajectory in basis.items()],
            start=Position(),
        )
        matrices = {
            axis: self.trajectory_matrix(trajectory, start).subs({n: 0})
            for axis, trajectory in basis.items()
        }
        return CMF(matrices, validate=False)

    @staticmethod
    def walk_symbol() -> sp.Symbol:
        return sp.Symbol("walk")

    def ctx(self, start: Position | None = None) -> FlintContext:
        start = Position(start) if start else Position()
        free_symbols = (
            self.free_symbols().union({CMF.walk_symbol()}).union(start.free_symbols())
        )
        return flint_ctx(free_symbols, fmpz=start.is_polynomial())

    def _calculate_diagonal_matrix_backtrack(
        self, trajectory: Position, start: Position, ctx: FlintContext
    ) -> SymbolicMatrix:
        """
        Inner function of an inner function. DO NOT USE DIRECTLY.
        This is the backtracking hook used for `_calculate_diagonal_marix`.
        It's used to look for a non-singular path towards the trajectroy matrix.
        """
        if trajectory.longest() == 0:
            return SymbolicMatrix.eye(self.rank(), ctx)
        for axis in sorted(trajectory.keys(), key=str):
            try:
                inner_trajectory = trajectory.copy()
                position = start.copy()
                sign = inner_trajectory[axis] >= 0
                current = SymbolicMatrix.from_sympy(self.M(axis, sign), ctx).subs(
                    position
                )
                position[axis] += inner_trajectory.pop(axis)
                return current * self._calculate_diagonal_matrix_backtrack(
                    inner_trajectory, position, ctx
                )
            except ZeroDivisionError:
                continue
        raise ZeroDivisionError(
            "A singularity has occured in every possible trajectory combination"
        )

    @lru_cache
    def _calculate_diagonal_matrix(
        self, trajectory: Position, start: Position, ctx: FlintContext
    ) -> SymbolicMatrix:
        """
        The manual calculation of trajectory matrix in the stopping condition.
        You should probably use `trajectory_matrix` instead.

        Assumes trajectory is a simple diagonal - all abs values are at most 1
        Args:
            trajectory: a dict containing the amount of steps in each direction.
            start: a dict representing the starting point of the multiplication.
        Returns:
            A matrix that represents a single step in the desired trajectory
        """
        if trajectory.longest() > 1:
            raise ValueError(
                f"Called _calculate_diagonal_matrix with a trajectory that is not a simple diagonal: {trajectory}"
            )
        trajectory = Position(
            {axis: value for axis, value in trajectory.items() if value != 0}
        )

        return self._calculate_diagonal_matrix_backtrack(trajectory, start, ctx)

    def _work_symbolic(
        self,
        trajectory: Position,
        start: Position,
    ) -> SymbolicMatrix:
        """
        Internal work logic. Do not use directly.
        """
        ctx = self.ctx(start)
        # Stopping condition: l-infinity norm of trajectory is less than 1
        if trajectory.longest() <= 1:
            return self._calculate_diagonal_matrix(trajectory, start, ctx)

        result = SymbolicMatrix.eye(self.rank(), ctx)
        symbol = CMF.walk_symbol()
        depth = trajectory.shortest()
        position = start.copy()
        while depth > 0:
            diagonal = trajectory.signs()
            result *= self._trajectory_matrix_inner(diagonal, position, symbol).walk(
                {symbol: 1}, int(depth), {symbol: 0}
            )
            position += depth * diagonal
            trajectory -= depth * diagonal
            depth = trajectory.shortest()
        return result

    def _work_numeric(
        self,
        trajectory: Position,
        start: Position,
    ) -> NumericMatrix:
        """
        Internal work logic. Do not use directly.
        """
        ctx = self.ctx(start)
        # Stopping condition: l-infinity norm of trajectory is less than 1
        if trajectory.longest() <= 1:
            matrix = self._calculate_diagonal_matrix(trajectory, start, ctx).factor()
            return NumericMatrix.lambda_from_rt(matrix)(start)

        result = NumericMatrix.eye(self.rank())
        symbol = CMF.walk_symbol()
        depth = trajectory.shortest()
        position = start.copy()
        while depth > 0:
            diagonal = trajectory.signs()
            result *= NumericMatrix.walk(
                self.trajectory_matrix(diagonal, position, symbol),
                Position({symbol: 1}),
                int(depth),
                Position({symbol: 0}),
            )
            position += depth * diagonal
            trajectory -= depth * diagonal
            depth = trajectory.shortest()
        return result

    def work(self, start: Position, end: Position) -> Matrix:
        """
        Returns the transformation matrix corresponding to walking from start to end
        """
        start = Position(start)
        end = Position(end)
        trajectory = end - start
        if not trajectory.is_integer():
            raise ValueError(
                f"Can only calculate work for two points with integer trajectory between them! "
                f"start={start}, end={end}, trajectory={trajectory}"
            )
        if start.free_symbols() == set():
            return self._work_numeric(trajectory, start).to_rt()
        else:
            return self._work_symbolic(trajectory, start).factor()

    def _trajectory_matrix_inner(
        self, trajectory: Position, start: Position, symbol: sp.Symbol
    ) -> SymbolicMatrix:
        start = self.trajectory_substitution(trajectory, start, symbol)
        return self._work_symbolic(trajectory, start)

    def trajectory_matrix(
        self, trajectory: dict, start: dict, symbol: sp.Symbol = n
    ) -> Matrix:
        """
        Returns a corresponding matrix for walking in a trajectory, up to a constant.
        If `start` is given, the new matrix will be reduced to a single variable `n`.
        Args:
            trajectory: a dict containing the amount of steps in each direction.
            start: a dict representing the starting point of the multiplication.
        Returns:
            A matrix that represents a single step in the desired trajectory
        Raises:
            ValueError: if trajectory, start and matrix keys do not match.
        """

        trajectory = Position(trajectory)
        start = Position(start)

        if self.axes() != trajectory.keys():
            raise ValueError(
                f"Trajectory axes {trajectory.keys()} do not match CMF axes {self.axes()}"
            )

        if not set(start.keys()).issubset(self.axes()):
            raise ValueError(
                f"Start axes {start.keys()} are not a subset of CMF axes {self.axes()}!"
            )

        return self._trajectory_matrix_inner(trajectory, start, symbol).factor()

    def trajectory_substitution(
        self, trajectory: Position, start: Position, symbol: sp.Symbol
    ) -> Position:
        """
        Returns the substitution of the start position to a function of the new trajectory symbol.

        This transformation is possible only when the starting point is known.
        Each incrementation of the variable `symbol` represents a full step in `trajectory`.

        Args:
            trajectory: The trajectory that was used to create the trajectory matrix.
            start: The starting point from which the walk operation is to be calculated.
            symbol: The new symbol of the reduced trajectory matrix.
        Returns:
            A dict representing the above variable reduction substitution
        Raises:
            ValueError: if trajectory keys and start keys do not match
        """
        trajectory = Position(trajectory)
        start = Position(start)

        if symbol in start or symbol in trajectory or symbol in start.free_symbols():
            raise ValueError(
                f"Attempted to run a trajectory_matrix where desired symbol is in start or trajectory! "
                f"symbol={symbol}, start={start}, trajectory={trajectory}"
            )
        effective_start = Position({axis: axis for axis in self.axes()})
        for axis in start:
            effective_start[axis] = start[axis]

        return effective_start + symbol * trajectory

    @batched("iterations")
    def walk(
        self,
        trajectory: dict,
        iterations: Batchable[int],
        start: dict,
    ) -> Batchable[Matrix]:
        r"""
        Returns a list of trajectorial walk multiplication matrices in the desired depths.

        The walk operation is defined as $\prod_{i=0}^{n-1}M_{s, t}(n)$,
        where M is `trajectory_matrix(trajectory, start)`, s is `start` and t is `trajectory`.

        Args:
            trajectory: A dict containing the amount of steps in each direction.
            iterations: The amount of trajectory matrix multiplications to perform, either an integer or a list.
            start: A dict representing the starting point of the multiplication.
        Returns:
            The limit of the walk multiplication as defined above.
            If `iterations` is a list, returns a list of limits.
        """
        if self.axes() != trajectory.keys():
            raise ValueError(
                f"Trajectory axes {trajectory.keys()} do not match CMF axes {self.axes()}"
            )

        if start and self.axes() != start.keys():
            raise ValueError(
                f"Start axes {start.keys()} do not match CMF axes {self.axes()}"
            )

        iterations_set = set(iterations)
        if len(iterations_set) != len(iterations):
            raise ValueError(f"`iterations` values must be unique, got {iterations}")

        if not iterations == sorted(iterations):
            raise ValueError(f"Iterations must be sorted, got {iterations}")

        if not all(depth >= 0 for depth in iterations):
            raise ValueError(
                f"iterations must contain only non-negative values, got {iterations}"
            )

        trajectory = Position(trajectory)
        start = Position(start)
        return self.trajectory_matrix(trajectory, start).walk(
            {n: 1}, iterations, {n: 0}
        )

    @batched("iterations")
    def limit(
        self,
        trajectory: dict,
        iterations: Batchable[int],
        start: dict,
        initial_values: Matrix | None = None,
        final_projection: Matrix | None = None,
    ) -> Batchable[Limit]:
        r"""
        Returns a list of limits of trajectorial walk multiplication matrices in the desired depths.

        The walk operation is defined as $\prod_{i=0}^{n-1}M(s_0 + i \cdot t_0, ..., s_k + i \cdot t_k)$,
        where `M=trajectory_matrix(trajectory, start)`, and `n / size(trajectory)` (L1 size - total amount of steps)

        Args:
            trajectory: A dict containing the amount of steps in each direction.
            iterations: The amount of trajectory matrix multiplications to perform, either an integer or a list.
            start: A dict representing the starting point of the multiplication.
            initial_values: the initial values matrix for the limit calculation
            final_projection: the final projection matrix for the limit calculation
        Returns:
            The limit of the walk multiplication as defined above.
            If `iterations` is a list, returns a list of limits.
        """

        def walk_function(iterations):
            return self.walk(trajectory, iterations, start)

        return Limit.walk_to_limit(
            iterations, walk_function, initial_values, final_projection
        )

    def delta(
        self,
        trajectory: dict,
        depth: int,
        start: dict,
        limit: float = None,
        initial_values: Matrix | None = None,
        final_projection: Matrix | None = None,
    ):
        r"""
        Calculates the irrationality measure $\delta$ defined, as:
        $|\frac{p_n}{q_n} - L| = \frac{1}{q_n}^{1+\delta}$
        for the walk specified by the trajectory.
        $p_n$ and $q_n$ are respectively defined as the [0,-1] and [1,-1] elements in walk's matrix (see `Limit` class).

        If limit is not specified (i.e, limit is None),
        then limit is approximated as limit = self.limit(2 * depth)

        Args:
            trajectory: the trajectory of the walk.
            depth: $n$, is the number of trajectory matrices multiplied.
                The $\ell_1$ distance walked from the start point is `depth * sum(trajectory.values())`.
            start: the starting point of the walk operation.
            limit: $L$
            initial_values: the initial values matrix for the limit calculation
            final_projection: the final projection matrix for the limit calculation
        Returns:
            the delta value as defined above.
        """
        if limit is None:
            m, mlim = self.limit(
                trajectory, [depth, 2 * depth], start, initial_values, final_projection
            )
            limit = mlim.as_float()
        else:
            m = self.limit(trajectory, depth, start, initial_values, final_projection)
        return m.delta(limit)

    def delta_sequence(
        self,
        trajectory: dict,
        depth: int,
        start: dict,
        limit: float = None,
        initial_values: Matrix | None = None,
        final_projection: Matrix | None = None,
    ):
        r"""
        Calculates delta values sequentially up to `depth`.

        Args:
            trajectory: the trajectory of the walk.
            depth: $n$, is the number of trajectory matrices multiplied.
                The $\ell_1$ distance walked from the start point is `depth * sum(trajectory.values())`.
            start: the starting point of the walk operation.
            limit: $L$
            initial_values: the initial values matrix for the limit calculation
            final_projection: the final projection matrix for the limit calculation
        Returns:
            the delta value as defined above.
        """
        depths = list(range(1, depth + 1))
        if limit is None:
            depths += [2 * depth]
            approximants = self.limit(
                trajectory, depths, start, initial_values, final_projection
            )
            limit = approximants[-1].as_float()
            approximants = approximants[:-1]
        else:
            approximants = self.limit(
                trajectory, depths, start, initial_values, final_projection
            )
        return [approximant.delta(limit) for approximant in approximants]



from .position import Position
from .matrix import Matrix
from .limit import Limit
from .generic_polynomial import GenericPolynomial
from .linear_recurrence import LinearRecurrence
from .simplify_object import simplify

__version__ = "0.0.1"

__all__ = [
    "Position",
    "Matrix",
    "Limit",
    "GenericPolynomial",
    "LinearRecurrence",
    "simplify",
]



from __future__ import annotations

from functools import cached_property
import copy
import itertools
from tqdm import tqdm

import sympy as sp
from sympy.abc import n

from ramanujantools import Matrix, Limit, GenericPolynomial
from ramanujantools.utils import batched, Batchable


def trim_trailing_zeros(sequence: list[int]) -> list[int]:
    ending = len(sequence)
    while ending > 0 and sequence[ending - 1] == 0:
        ending -= 1
    return sequence[0:ending]


class LinearRecurrence:
    r"""
    Represents a linear recurrence of the form
    $\sum_{i=0}^{N}a_i(n) p(n - i) = 0$

    (Note: equivalent to $-a_0(n)p(n) = \sum_{i=1}^{N}a_i(n) p(n - i)$)

    Note that the beginning index can be decided later,
    therefore this class can represents all recurrences of the form
    $\sum_{i=0}^{N}a_i(n + s) p(n - i) = 0$ for any integer $s$.
    """

    def __init__(self, recurrence: Matrix | list[sp.Expr] | None = None):
        r"""
        Construct the recurrence.

        The recurrence argument can be one of two types:
            1. A list of the coefficients of the recurrence [a_0(n), ..., a_N(n)]
            2. A matrix which is companionized and used as the recurrence sequence
        """
        if recurrence is None:
            relation = []
        elif isinstance(recurrence, Matrix):
            recurrence_matrix = recurrence.as_companion()
            col = recurrence_matrix.col(-1)
            lead = col.denominator_lcm
            coeffs = [sp.simplify(p * lead) for p in reversed(col)]
            relation = [-lead] + coeffs
        else:
            relation = recurrence
        relation = [sp.factor(sp.simplify(p)) for p in relation]
        self.relation = trim_trailing_zeros(relation)

    def __eq__(self, other: Matrix) -> bool:
        """
        Returns True iff two requrences are identical (even up to gcd).
        """
        return self.relation == other.relation

    def __neg__(self) -> LinearRecurrence:
        return LinearRecurrence([-c for c in self.relation])

    def __add__(self, other: LinearRecurrence) -> LinearRecurrence:
        return LinearRecurrence(
            [
                a_i + b_i
                for a_i, b_i in itertools.zip_longest(
                    self.relation, other.relation, fillvalue=0
                )
            ]
        )

    def __radd__(self, other: LinearRecurrence) -> LinearRecurrence:
        return self + other

    def _shift(self, num: int) -> LinearRecurrence:
        return LinearRecurrence(
            [0] * num + [c.subs({n: n - num}) for c in self.relation]
        )

    def __sub__(self, other: LinearRecurrence) -> LinearRecurrence:
        return self + (-other)

    def __rsub__(self, other: LinearRecurrence) -> LinearRecurrence:
        return self - other

    def __mul__(self, scalar: int) -> LinearRecurrence:
        return LinearRecurrence([p * scalar for p in self.relation])

    def __rmul__(self, scalar: int) -> LinearRecurrence:
        return self * scalar

    def __truediv__(self, scalar: int) -> LinearRecurrence:
        return LinearRecurrence([p / scalar for p in self.relation])

    def __floordiv__(self, scalar: int) -> LinearRecurrence:
        return LinearRecurrence([p / scalar for p in self.relation])

    def __repr__(self) -> str:
        return f"LinearRecurrence({self.relation})"

    def __str__(self) -> str:
        return f"{self._symbolic_relation()} = 0"

    def _latex(self, printer) -> str:
        return f"{printer.doprint(self._symbolic_relation())} = 0"

    def _repr_latex_(self) -> str:
        return rf"{sp.latex(self)}"

    def _symbolic_relation(self) -> sp.Expr:
        terms = [
            self.relation[i] * sp.Function("p")(n - i)
            for i in range(len(self.relation))
        ]
        return sp.Add(*terms)

    @cached_property
    def gcd(self) -> sp.Expr:
        """
        Returns the GCD of all recurrence coefficients
        """
        return sp.gcd([r.as_numer_denom()[0] for r in self.relation])

    @cached_property
    def denominator_lcm(self) -> sp.Expr:
        return sp.lcm([r.as_numer_denom()[1] for r in self.relation])

    @cached_property
    def recurrence_matrix(self) -> Matrix:
        """
        Returns the companion form recurrence matrix corresponding to the recurrence
        """
        denominator = sp.simplify(-self.relation[0])
        column = [c / denominator for c in self.relation[1:]]
        return Matrix.companion_form(list(reversed(column)))

    def order(self) -> int:
        """
        Returns the order of the recurrence
        """
        return len(self.relation) - 1

    def degrees(self) -> list[int]:
        """
        Returns a list of the degrees of all coefficients
        """
        return [sp.Poly(p, n).degree() for p in self.relation]

    def subs(self, substitutions: dict[sp.Symbol, sp.Expr]) -> LinearRecurrence:
        """
        Substitutes symbols in the recurrence.
        """
        return LinearRecurrence([p.subs(substitutions) for p in self.relation])

    def free_symbols(self) -> set[sp.Symbol]:
        """
        Returns all free symbols of the recurrence (including `n`)
        """
        return set.union(*[p.free_symbols for p in self.relation])

    def parameters(self) -> set[sp.Symbol]:
        """
        Returns all symbolic parameters of the recurrence (excluding `n`)
        """
        return self.free_symbols() - self.axes()

    def normalize(self) -> LinearRecurrence:
        """
        Normalizes the recurrence, setting the leading coefficient to 1 by inflating by it.
        """
        return self.inflate(-self.relation[0]).simplify()

    def simplify(self) -> LinearRecurrence:
        """
        Simplifies the coefficients of the recurrence
        """
        relation = [p * self.denominator_lcm / self.gcd for p in self.relation]
        return LinearRecurrence([sp.factor(p.simplify()) for p in relation])

    @batched("iterations")
    def limit(
        self, iterations: Batchable[int], start=0, initial_values: Matrix = None
    ) -> Batchable[Limit]:
        r"""
        Returns the Limit matrix of the recursion up to a certain depth
        """
        return self.recurrence_matrix.limit(
            {n: 1}, iterations, {n: start}, initial_values
        )

    def evaluate_solution(
        self, initial_values: Matrix, start: int, end: int
    ) -> list[sp.Rational]:
        """
        Returns an evaluation of a specific solution of the recurrence.
        A specific solution is uniquely defined by initial values.
        Args:
            initial_values: A row matrix (1xN) of the initial values of the solution.
            inidices: The solution indices required to evaluate.
            given_index: The highest index of the inital values.
        Returns:
            A list of evaluated points of the specific recurrence
        """
        if self.order() != len(initial_values):
            raise ValueError(
                "Initial values of a recursion must be of the recurrence's order! "
                f"got {len(initial_values)} while order is {self.order()}"
            )
        if not start <= end:
            raise ValueError("Requested to evaluate solution at a negative range!")
        retval = []
        iterations = list(range(1, end - start + 1))
        limits = self.limit(
            iterations, start, Matrix.vstack(initial_values, initial_values)
        )
        for limit in limits:
            retval.append(limit.p())
        return retval

    def inflate(self, c: sp.Expr) -> LinearRecurrence:
        r"""
        Inflates the recurrence by a polynomial c.

        The inflated recurrence satisfies
        $\sum_{i=0}^{N}\left(\prod_{j=0}^{i-1}c(n-i)\right)a_i(n) p(n - i) = 0$

        The inflated recurrence converges to the same limit up to different initial values.
        """
        c = sp.simplify(c).as_expr()
        current = c
        relation = copy.deepcopy(self.relation)
        for i in range(1, len(self.relation)):
            relation[i] *= current
            current *= c.subs({n: n - i})
        return LinearRecurrence(relation)

    def deflate(self, c: sp.Expr) -> LinearRecurrence:
        r"""
        Deflates the recurrence by a polynomial c.

        Equivalent to `self.inflate(1 / c)`.
        """
        recurrence = self.inflate(1 / c)
        return recurrence

    def fold(self, multiplier: sp.Expr) -> LinearRecurrence:
        r"""
        Folds the recurrence into a higher order recurrence.

        Given a recurrence
        $$H_n := \sum_{i=0}^{N}a_i(n) p(n - i) = 0.$$

        We rewrite $n \to n-1$ to get:
        $$H_{n-1} := \sum_{i=0}^{N}a_i(n-1) p(n - 1 - i) = 0.$$

        Selecting a multiplier rational function $d(n)$, this function returns $H_n + d(n) H_{n-1} = 0$.
        Note: this is the same as `LinearRecurrence([1, multiplier]).compose(self)`.

        Example:
            >>> s = LinearRecurrence([sp.Function("a")(n), sp.Function("b")(n), sp.Function("c")(n)])
            >>> s
            LinearRecurrence([a(n), b(n), c(n)])
            >>> s.fold(sp.Function("d")(n))
            LinearRecurrence([a(n), a(n - 1)*d(n) + b(n), b(n - 1)*d(n) + c(n), c(n - 1)*d(n)])
        """
        return self + multiplier * self._shift(1)

    @staticmethod
    def all_divisors(p: sp.Poly) -> list[sp.Poly]:
        r"""
        Returns all divisors of polynomial `p`.
        Assumes p is polynomial in `n`.
        """
        p = sp.Poly(p, n)
        content, factors_list = p.factor_list()
        factors = []
        for factor, order in factors_list:
            factors.append([factor**d for d in range(order + 1)])
        if len(content.free_symbols) == 0:
            for root, order in content.factors().items():
                factors.append([root**d for d in range(order + 1)])
        combinations = itertools.product(*factors)
        divisors = []
        for combination in combinations:
            divisors.append(sp.prod(combination))
        return divisors

    def possible_multipliers(self) -> list[sp.Poly]:
        r"""
        Returns all candidates for a multiplier rational $d(n)$
        that could have been used to fold a lesser order recursion into this one.
        """
        return LinearRecurrence.all_divisors(self.relation[-1])

    def unfold_poly(self, multiplier: sp.Poly) -> list[LinearRecurrence]:
        r"""
        Attempts to unfold this recursion using a multiplier rational function.
        In case of success, returns all recurrence that satisfy `recurrence.fold(multiplier) == self`

        If `self` contains parameters (other than n), will attempt to find matching substitutions.
        In case of success, the returned recurrences will be substituted with the solution.
        i.e, for that solution, `recurrence.fold(multiplier) == self.subs(solution)`
        """
        multiplier = sp.Poly(multiplier, n).as_expr()
        unfolded = [self.relation[0]]
        for index in range(1, len(self.relation) - 1):
            next = unfolded[index - 1].subs({n: n - 1})
            unfolded.append(sp.simplify(self.relation[index] - next * multiplier))
        expected_tail = unfolded[-1].subs({n: n - 1}) * multiplier
        if sp.simplify(expected_tail - self.relation[-1]) == 0:
            return [LinearRecurrence(unfolded)]
        if len(self.free_symbols()) > 0:
            solutions = sp.solve(sp.Poly(expected_tail - self.relation[-1], n).coeffs())
            return [LinearRecurrence(unfolded).subs(solution) for solution in solutions]
        return None

    def unfold(self, inflation_degree=0) -> tuple[LinearRecurrence, sp.Poly]:
        r"""
        Attempts to unfold this recursion by enumerating over all possible
        multiplier candidates and attempting to unfold using them.

        If `inflation_degree` is not zero, will inflate `self` by a generic polynomial of that degree.
        The solver will then attempt to find a solution for that inflation that manages to unfold the recurrence.
        """
        results = []
        if inflation_degree > 0:
            inflation, _ = GenericPolynomial.of_degree(inflation_degree, "c", n)
            recurrence = self.inflate(inflation)
        elif inflation_degree < 0:
            inflation, _ = GenericPolynomial.of_degree(-inflation_degree, "c", n)
            recurrence = self.deflate(inflation).normalize()
        else:
            recurrence = self
        for multiplier in tqdm(self.possible_multipliers()):
            solutions = recurrence.unfold_poly(multiplier)
            for unfolded in solutions:
                results.append((unfolded, multiplier))
        return results

    def compose(self, other: LinearRecurrence) -> LinearRecurrence:
        r"""
        Composes two linear recurrences.
        Given two linear recurrence $A_n := \sum_{i=0}^{d_a}a_i(n) p(n - i) = 0$
        and $B_n := \sum_{i=0}^{d_b}b_i(n) p(n - i) = 0$, calculates the composition of them:

        $$ A_n \circ B_n := \sum_{i=0}^{d_a} a_i(n) \cdot B_{n-i} = 0 $$

        The resulting recurrence is of order $d_a + d_b$.
        """
        result = LinearRecurrence()
        for i, a_i in enumerate(self.relation):
            result += a_i * other._shift(i)
        return result

    def kamidelta(self, depth=20):
        r"""
        Uses the Kamidelta alogrithm to predict possible delta values of the recurrence.
        Effectively calls kamidelta on `recurrence_matrix`.

        For more details, see `Matrix.kamidelta`
        """
        return self.recurrence_matrix.kamidelta(depth)



from pytest import approx

import itertools

import sympy as sp
from sympy.abc import x, y, n

from ramanujantools.pcf import PCF
from ramanujantools.cmf import known_cmfs


def test_cmf_e():
    cmf = known_cmfs.e()
    limit = cmf.limit({x: 1, y: 1}, 50, {x: 0, y: 0})
    assert limit.as_float() == approx(1 / (1 - limit.mp.e))


def test_cmf_pi():
    cmf = known_cmfs.pi()
    limit = cmf.limit({x: 1, y: 1}, 50, {x: 1, y: 0})
    assert limit.as_float() == approx((2 - limit.mp.pi) / 2)


def test_cmf_zeta3():
    cmf = known_cmfs.zeta3()
    limit = cmf.limit({x: 1, y: 1}, 50, {x: 1, y: 1})
    assert limit.as_float() == approx((1 - limit.mp.zeta(3)) / limit.mp.zeta(3))


def test_apery():
    cmf = known_cmfs.zeta3()
    pcf = PCF(cmf.trajectory_matrix({x: 1, y: 1}, {x: 0, y: 0})).deflate_all()
    # This is Apery's PCF
    assert pcf == PCF(34 * n**3 + 51 * n**2 + 27 * n + 5, -(n**6))

    depth = 2000
    actual_limit = pcf.limit(depth)
    ctx = actual_limit.mp
    expected_limit = ctx.mpf(6 / ctx.zeta(3))
    assert actual_limit.as_float() == approx(float(expected_limit))
    delta = pcf.delta(depth, expected_limit)
    assert delta > 0.08


def test_cmf1():
    c0, c1, c2, c3 = sp.symbols("c:4")

    cmf = known_cmfs.cmf1()
    for a, b in itertools.product(range(1, 10), range(1, 10)):
        limit = cmf.subs({c0: 0, c1: a, c2: 0, c3: b}).limit(
            {x: 1, y: 1}, 50, {x: 1, y: 1}
        )
        assert limit.as_float() == approx(-a + b / limit.mp.log(1 + b / a), 1e-4)


def test_all_conserving():
    r"""
    Checks that all cmfs are indeed cmfs.
    If one of these is not a cmf, an exception will be thrown.
    """
    known_cmfs.e().validate_conserving()
    known_cmfs.pi().validate_conserving()
    known_cmfs.symmetric_pi().validate_conserving()
    known_cmfs.zeta3().validate_conserving()
    known_cmfs.var_root_cmf().validate_conserving()
    known_cmfs.cmf1().validate_conserving()
    known_cmfs.cmf2().validate_conserving()
    known_cmfs.cmf3_1().validate_conserving()
    known_cmfs.cmf3_2().validate_conserving()
    known_cmfs.cmf3_3().validate_conserving()
    known_cmfs.hypergeometric_derived_2F1().validate_conserving()
    known_cmfs.hypergeometric_derived_3F2().validate_conserving()



import pytest

from ramanujantools.utils import batched, Batchable


class Dummy:
    @batched("x")
    def process(self, x: Batchable[int]) -> Batchable[int]:
        return [i * 2 for i in x]


def test_scalar_input_returns_scalar():
    d = Dummy()
    result = d.process(5)
    assert result == 10  # single scalar return, unwrapped


def test_list_input_returns_list():
    d = Dummy()
    result = d.process([1, 2, 3])
    assert result == [2, 4, 6]


def test_invalid_scalar_type_raises():
    d = Dummy()
    with pytest.raises(TypeError):
        d.process("string_instead_of_int")


def test_invalid_list_element_type_raises():
    d = Dummy()
    with pytest.raises(TypeError):
        d.process([1, "bad", 3])


def test_missing_argument_raises():
    d = Dummy()
    with pytest.raises(TypeError):
        Dummy.process(d)  # no argument


def test_preserves_other_args():
    class Dummy2:
        @batched("y")
        def f(self, x: int, y: Batchable[int]) -> Batchable[int]:
            return [x + i for i in y]

    d = Dummy2()
    assert d.f(10, 5) == 15
    assert d.f(10, [1, 2]) == [11, 12]


def test_missing_annotation_raises():
    with pytest.raises(TypeError, match="No annotation for argument 'missing'"):

        class Bad:
            @batched("missing")
            def foo(self, x: int) -> None:
                pass


def test_invalid_annotation_format_raises():
    with pytest.raises(TypeError, match=r"annotation must be.*List"):

        class Bad2:
            @batched("x")
            def foo(self, x: int) -> None:
                pass


def test_argument_not_provided_raises():
    d = Dummy()
    # Cannot skip required argument without triggering Python's own TypeError,
    # so just test that TypeError (not our ValueError) is raised:
    with pytest.raises(TypeError):
        d.process()


def test_scalar_wrong_type_raises():
    d = Dummy()
    with pytest.raises(TypeError, match="Argument 'x' must be of type Batchable"):
        d.process("not an int")


def test_list_element_wrong_type_raises():
    d = Dummy()
    with pytest.raises(
        TypeError, match="All elements of argument 'x' must be of type Batchable"
    ):
        d.process([1, "bad", 3])



from .context import (
    flint_ctx,
    flint_from_sympy,
    flint_to_sympy,
    FlintPoly,
    FlintContext,
)
from .rational import FlintRational
from .symbolic_matrix import SymbolicMatrix
from .numeric_matrix import NumericMatrix

__all__ = [
    "flint_ctx",
    "flint_from_sympy",
    "flint_to_sympy",
    "FlintPoly",
    "FlintRational",
    "FlintContext",
    "SymbolicMatrix",
    "NumericMatrix",
]



from __future__ import annotations

import math
import flint
import sympy as sp

from ramanujantools import Position
from ramanujantools.flint_core import (
    FlintPoly,
    FlintContext,
    flint_from_sympy,
    flint_to_sympy,
)


class FlintRational:
    """
    Represents a rational function.
    Implemented as a numerator and denominator, reduces by gcd every step
    """

    def __init__(
        self, numerator: FlintPoly, denominator: FlintPoly, ctx: FlintContext
    ) -> FlintRational:
        self.is_integer = isinstance(numerator, flint.fmpz_mpoly)
        gcd = numerator.gcd(denominator)
        if not self.is_integer:
            content = FlintRational.fmpq_gcd(numerator.coeffs() + denominator.coeffs())
            gcd *= content
        self.numerator = numerator / gcd
        self.denominator = denominator / gcd
        self.ctx = ctx

    @staticmethod
    def from_sympy(rational: sp.Expr, ctx: FlintContext) -> FlintRational:
        r"""
        Converts a rational function given as a sympy expression to a FlintRational.
        Args:
            rational: The expression to convert to flint
            ctx: The desired mpoly context (which also defines the supported variables)
        Returns:
            A FlintRational object representing the `rational` value
        """
        numerator, denominator = rational.as_numer_denom()
        return FlintRational(
            flint_from_sympy(numerator, ctx),
            flint_from_sympy(denominator, ctx),
            ctx,
        )

    @staticmethod
    def fmpq_gcd(numbers: list[flint.fmpq]) -> flint.fmpz:
        denominator = flint.fmpz(1)
        for c in numbers:
            denominator *= c.denominator
        numerators = [c.numerator * denominator / c.denominator for c in numbers]
        gcd = math.gcd(*numerators)
        return flint.fmpq(gcd, denominator)

    def inv(self) -> FlintRational:
        """
        Returns 1 / self.
        """
        return FlintRational(self.denominator, self.numerator, self.ctx)

    def __neg__(self):
        return FlintRational(-self.numerator, self.denominator, self.ctx)

    def __add__(self, other: FlintRational) -> FlintRational:
        return FlintRational(
            self.numerator * other.denominator + self.denominator * other.numerator,
            self.denominator * other.denominator,
            self.ctx,
        )

    def __radd__(self, other: FlintRational) -> FlintRational:
        return self + other

    def __sub__(self, other) -> FlintRational:
        return self + (-other)

    def __rsub__(self, other) -> FlintRational:
        return -self + other

    def __mul__(self, other) -> FlintRational:
        if isinstance(other, FlintRational):
            numerator = self.numerator * other.numerator
            denominator = self.denominator * other.denominator
            return FlintRational(numerator, denominator, self.ctx)
        else:
            return FlintRational(self.numerator * other, self.denominator, self.ctx)

    def __rmul__(self, other) -> FlintRational:
        return self * other

    def __truediv__(self, other) -> FlintRational:
        if isinstance(other, FlintRational):
            return self * other.inv()
        return FlintRational(self.numerator, self.denominator * other, self.ctx)

    def __rtruediv__(self, other) -> FlintRational:
        return other * self.inv()

    def __repr__(self) -> str:
        return f"FlintRational({self.numerator}, {self.denominator})"

    def __eq__(self, other: FlintRational) -> bool:
        return (
            self.numerator == other.numerator and self.denominator == other.denominator
        )

    def degrees(self) -> list[int]:
        return [max(poly.degrees()) for poly in [self.numerator, self.denominator]]

    def subs(self, substitutions: dict) -> FlintRational:
        """
        Substitutes symbols in self.
        """
        substitutions = Position(
            {str(key): value for key, value in substitutions.items()}
        )
        composition = []
        for gen in self.ctx.gens():
            if str(gen) in substitutions:
                value = flint_from_sympy(substitutions[str(gen)], self.ctx)
            else:
                value = gen
            composition.append(value)
        content = (
            1
            if self.is_integer
            else substitutions.denominator_lcm() ** max(self.degrees())
        )
        return FlintRational(
            (content * self.numerator).compose(*composition),
            (content * self.denominator).compose(*composition),
            self.ctx,
        )

    def factor(self) -> sp.Expr:
        """
        Factors self and returns it as a sp.Expr
        """
        return flint_to_sympy(self.numerator) / flint_to_sympy(self.denominator)



from pytest import approx, raises
import sympy as sp
from sympy.abc import n, z

from ramanujantools.pcf import (
    PCF,
    HypergeometricLimit,
    Hypergeometric1F1Limit,
    Hypergeometric2F1Limit,
)


def test_factory_bad_degrees_throw():
    with raises(ValueError):
        HypergeometricLimit(PCF(0, n))
    with raises(ValueError):
        HypergeometricLimit(PCF(n, 0))
    with raises(ValueError):
        HypergeometricLimit(PCF(n**2, n))
    with raises(ValueError):
        HypergeometricLimit(PCF(n, n**3))


def test_factory_types():
    assert isinstance(HypergeometricLimit(PCF(n, n)), Hypergeometric1F1Limit)
    assert isinstance(HypergeometricLimit(PCF(n, n**2)), Hypergeometric2F1Limit)


def test_1f1_limit():
    pcf = PCF(n + 1, n + 1)
    limit = sp.E - 1
    assert sp.simplify(limit - Hypergeometric1F1Limit(pcf).limit()) == 0


def test_1f1_limit_parametric():
    pcf = PCF(n + 2 - z, z * (n + 1))
    limit = (z * (sp.exp(z) - 1)) / ((z - 1) * sp.exp(z) + 1)
    assert sp.simplify(limit - Hypergeometric1F1Limit(pcf).limit()) == 0


def test_2f1_limit():
    pcf = PCF(5 + 10 * n, 1 - 9 * n**2)
    limit = (sp.root(4, 3) + 1) / (sp.root(4, 3) - 1)
    assert float(limit - Hypergeometric2F1Limit(pcf).limit()) < 10e-50


def test_2f1_limit_parameteric():
    # going for positive c here to avoid piecewise condition of c!=-2
    c = sp.Symbol("c", positive=True)
    pcf = PCF((1 + 2 * n) * (c + 2), 1 - (c * n) ** 2)
    # limit = (sp.root(c + 1, c) + 1) / (sp.root(c + 1, c) - 1)
    hyp = Hypergeometric2F1Limit(pcf)
    assert hyp.alpha1 == -1 / c
    assert hyp.alpha2 == 1 / c
    assert hyp.beta == sp.Rational(1, 2)
    assert hyp.z == (-c + 2 * sp.sqrt(c + 1) - 2) / (4 * sp.sqrt(c + 1))


def test_1f1_limit_numerically():
    pcf = PCF(1 + 2 * n, 3 + 4 * n)
    assert pcf.limit(1000).as_float() == approx(
        float(Hypergeometric1F1Limit(pcf).limit())
    )


def test_2f1_limit_numerically():
    pcf = PCF(5 + 10 * n, 1 - 9 * n**2)
    assert pcf.limit(1000).as_float() == approx(
        float(Hypergeometric2F1Limit(pcf).limit())
    )


def test_1f1_subs():
    from sympy.abc import c

    pcf = PCF(c * n, 3 - n)
    hyp = Hypergeometric1F1Limit(pcf)
    assert hyp.subs(c, 7) == Hypergeometric1F1Limit(pcf.subs(c, 7))


def test_2f1_subs():
    from sympy.abc import c

    pcf = PCF((1 + 2 * n) * (c + 2), 1 - (c * n) ** 2)
    hyp = Hypergeometric2F1Limit(pcf)
    assert hyp.subs(c, 7) == Hypergeometric2F1Limit(pcf.subs(c, 7))
